/****** Segwii Code   ******
***     Version 1.0      ***
**** By: Ara Kourchians ****
***************************/


int X_axis = 0;                //Input pin of x-axis                   (analog pin)
int X_gyro_axis = 3;           //Input pin of x-axis gyro input        (analog pin)
int Vref = 6;                  //Input pin of the Voltage reference    (analog pin)

double x_acc_shift = 397;      //Sets the shift on the x-axis accelerometer reading     ******************* set and adjust    NOTE: Each sensor is different   (AKA You need to set it for your own hardware!)
double x_acc_scale = 300;      //Sets the scale on the x-axis accelerometer reading     ******************* set and adjust    NOTE: Each sensor is different   (AKA You need to set it for your own hardware!)

double x_gyro_shift = 305;     //Sets the shift on the x-axis gyro reading              ******************* set and adjust    NOTE: Each sensor is different   (AKA You need to set it for your own hardware!)
double x_gyro_scale = 114;     //Sets the scale on the x-axis accelerometer reading     ******************* set and adjust    NOTE: Each sensor is different   (AKA You need to set it for your own hardware!)
  
double Volt_ref;               //Voltage Reference  (millivolts)
  
double mV_acc;                 //millvolt reading from the accelerometer
double dmV_acc;                //delta millivolts (difference) reading from accelerometer on the 5DoS IMU  (Shifted by x_acc_shift)
double x_raw_angle = 0;        //Raw angle read from the accelerometer 5DoS IMU (no filter)

double mV_gyro;                //millivolt reading from the gyroscope

double x_angle_nf;             //X angle no filter                (degrees)
double x_angle = 0;            //Clean Signal X-axis Angle        (degrees)
double x_ang_vel;              //Clean Signal X-axis Angular Rate (degrees)

double dt = 0.052;             //************************************************ milliseconds to do a cycle of the loop  (measure with o-scope)   52Hz loop

double cmd_angle = 0.0;        //Commanded Angle      (default is 0 degrees so the robot stays upright)
double ref_angle;              //Reference Angle
double angle_error;            //Error of the angle   (relative to commanded and reference angle)
double old_angle_error = 0.0;  //Previous Angle Error (Initially 0 because an initial term is need to compair errors)
double Pterm;                  //Proportional Term    This control changes torque based on present information
double Iterm = 0.0;            //Integral Term        This control changes torque based on past information                      (Initially 0 because an initial term is need to integrate)
double Dterm;                  //Derivative Term      This control changes torque based on predicting future information
double AngVelTerm;             //AngVelTerm Term      This control compensates for any angular velocity
double Pgain = 80.0;           //Proportional Constant        NOTE: This varies based on motor specs, and physical conditions    (AKA You need to set it for your own hardware!)
double Igain = 100.0;          //Integral Constant            NOTE: This varies based on motor specs, and physical conditions    (AKA You need to set it for your own hardware!)
double Dgain = 0.00001;        //Derivative Contstant         NOTE: This varies based on motor specs, and physical conditions    (AKA You need to set it for your own hardware!)
double Fgain = 5.00;           //AngVelTerm Constant          NOTE: This varies based on motor specs, and physical conditions    (AKA You need to set it for your own hardware!)


double torque;                 //Torque required to compensate

int switchPin = 0;             //switchPin: Digital pin 0 - 50 : This sets it either back or forward
int motorPin = 5;              //motorPin: Analog pin 0 - 5 : This gives the amount of power to the motor 0 - 1024

int input_data = 2000;         //Input Data from Processing



//******* Voltage Reference *******
double getVref()
{
  Volt_ref = (double)((analogRead(Vref)/229.0)*1000);                         //Converts bits read to millivolts (bit_reading/229)*1000
}
//******* Voltage Reference End *******

//******* Accelerometer Drivers 5DoS IMU ******  
double getAcc()
{
  mV_acc = (double)((analogRead(X_axis)/229.0)*1000);                         //Converts bits read to millivolts (bit_reading/229)*1000
  getVref();
  mV_acc = mV_acc - Volt_ref;
  dmV_acc = (mV_acc-x_acc_shift);        
  
    if(dmV_acc >= -250 && dmV_acc <= 250)
    {
        x_raw_angle = (double)((asin(dmV_acc/x_acc_scale))*(180.0/3.141592));
        return x_raw_angle;                                                   //Converts mv input to angle position in Degrees (Theta)
    }
    else
    {
        return x_raw_angle;                                                   //Returns raw angle if accelerometer is reading anything larger than 1g
    }
}
//******* End Accelerometer Drivers 5DoS IMU ******


//******* Gyro Sensor Drivers 5DoS IMU ******
double getGyro()
{               
  double gyro_bit = analogRead(X_gyro_axis);
  mV_gyro = (double)((gyro_bit/229.0)*1000);                                                      //Converts bits read to millivolts by diving by 229 and * by 1000
  getVref();
  mV_gyro = mV_gyro - Volt_ref;
  return (double)(-1.0*(((mV_gyro - x_gyro_shift)/x_gyro_scale)*(180.0/3.141592)));               //Returns Degrees/Second
}
//******* End Gyro Sensor Drivers 5DoS IMU ******


//******** Motor Driver ********     ***** CURRENTLY UNIDRIVE ******
//input_value: -1000 - 1000
void motorDrive(int input_value) 
{
  if(input_value >= 0) 
  {
    digitalWrite(switchPin, HIGH);                       //Signals H-Bridge to move power forward
    analogWrite(motorPin, input_value);
  }
  else if(input_value < 0) 
  { 
    input_value = abs(input_value);
    digitalWrite(switchPin, LOW);                        //Signals H-Bridge to move power backward
    analogWrite(motorPin, input_value);
  }
}
//******** End of Motor Driver ********    ***** CURRENTLY UNIDRIVE ******




//******* Display Mode ******
//Serial Display Output (USB)
void display()
{
  Serial.print((int)(x_angle_nf));       //No Filter
  Serial.print(" ");
  Serial.print((int)(x_ang_vel));        //Angluar Velocity
  Serial.print(" ");
  Serial.print((int)(x_angle));          //Filtered angle
  Serial.print("         ");
  Serial.print((int)(mV_acc));           //Millivolt reading from the acceleromter
  Serial.print(" ");
  Serial.print((int)(mV_gyro));          //Millivolt reading from the gyroscope
  Serial.print("         ");
  Serial.print((int)(torque));           //Torque being sent to the motor
  Serial.print("         ");
  Serial.print((int)(Pterm));              //Proportional Term
  Serial.print("  ");
  Serial.print((int)(Iterm));              //Integral Term
  Serial.print("  ");
  Serial.print((int)(Dterm));              //Derivative Term
  Serial.print("  ");
  Serial.print((int)(cmd_angle));         //Data input Term
  Serial.println("  ");
  delay(40);
}
//******* Display Mode ******



void setup()
{
  pinMode(switchPin, OUTPUT);           //HIGH_LOW port
  pinMode(motorPin, OUTPUT);            //Power gates to motor
  pinMode(1, OUTPUT);                   //Enable H-Bridge cable
  pinMode(48, OUTPUT);                  //Status Light
  pinMode(16, OUTPUT);                  //Red Light
  pinMode(17, OUTPUT);                  //Yellow Light  
  pinMode(18, OUTPUT);                  //Green Light
  
  pinMode(24, INPUT);                   //Enables the accelerometer input
  
  for(int i=0;i<10;i++)
  {
    digitalWrite(16, HIGH);             //(Red) Light  (Tells user to prepare, the segwii is about to start)
    delay(100);
    digitalWrite(16, LOW);
    delay(100);
  }
  
  digitalWrite(17, HIGH);               //(Yellow) Light  (Tells the user to really prepair... the segwii is 800ms from starting!)
  delay(800); 
  digitalWrite(17, LOW);                //(Yellow) Light
  
  Serial.begin(115200);                 //BAUD rate set to 115200
  delay(100);                           
}


void loop()
{
  digitalWrite(48, HIGH);                                                            //Light indicates process has begun
  digitalWrite(18, HIGH);                                                            //Operation Light (Green)
  digitalWrite(1, HIGH);                                                             //Enables H-Bridge


  //*** Input and Complementary Filter tilt angle and gyro ***
    
    x_angle_nf = getAcc();                                                           //Calls the acc drivers to get angle (Theta) (Degrees)
    
    x_ang_vel = getGyro();                                                           //Calls the gyro drivers to get angular velocity (Omega) (Degrees/Second)
    
    x_angle = ((0.9)*((x_angle) + (x_ang_vel*dt)) + (0.1)*(x_angle_nf));             //Low-Pass Filter Accelerometer - High-Pass Filter Gyro : Output is a clean x-axis angle position
    
 //*** End Input and Complementary Filter tilt angle and gyro ***


 //*** Takes input from Wiimote *** 
   if(Serial.available())                                                            //Is there a serial input available?
   { 
      input_data = Serial.read();                                                    //A-Ha! Yes there is! Here is the input from the Wiimote
   }   

   if(input_data >= -1000 && input_data <=1000)                                      //Throws out any commanded angle that is too large
   {
     cmd_angle = input_data/10;                                                      //Converts the input data to a commanded angle
     
     if(cmd_angle >= 2)                                                              //Caps the maximum commanded angle to 2 degrees
     {   
        cmd_angle = 2;
     }
     if(cmd_angle <= -2)                                                             //Caps the maximum commanded angle to 2 degrees
     {
        cmd_angle = -2;
     }  
   }
 //*** End of input from Wiimote ***
  
  
 //*** Motor Control (Torque calculation and PID) ***                                      
    ref_angle = x_angle;                                                             //Sets the reference angle as the current angle

    angle_error = (-1.0)*(cmd_angle - ref_angle);                                    //Determine angle error by subtracting the commanded angle by the reference  (negative one is to invert the effect the minus makes)

    Pterm = angle_error * Pgain;                                                     //Determine the Proportional Term by multiplying your Proportional gain to the error
    Iterm = (Iterm + (angle_error * Igain)) * .9999;                                 //Determine the Integral Term by integrating angle error with a slow washout to zero
    Dterm = (angle_error - old_angle_error) * Dgain;                                 //Determine the Derivative Term by subtracting the current error from the old error then multiply to scale it
    AngVelTerm = x_ang_vel * Fgain;                                                  //AngVelTerm is the compensation for any angular velocity, this will counter act it

    torque = Pterm + Dterm + Iterm + AngVelTerm;                                     //Sums up the PID to give a final torque

    old_angle_error = angle_error;                                                   //Takes the current angle and makes it the old angle
    


    if(torque >= 1000)                                                               //Limits the max speed
    {
      torque = 1000;                                                                 //Caps output torque to 1000
    }
    if(torque <= -1000)                                                              //Limits the max speed
    {
      torque = -1000;                                                                //Caps output torque to 1000
    } 

    motorDrive(torque);                                                              //Sends final torque to motors
 //*** End Motor Control (Torque calculation and PID) *** 
 
 
  display();                                                                         //Allows Serial Communication and Diagnosis
  digitalWrite(18, LOW);                                                             //Blinks off Green, Using an Osciliscope you can measure the processor timing
  delay(10);                                                                         //Time to catch a breath
}


/*
    So long and thanks for all the fish!
    
                            _ 
                          ,' \ 
                        ,'    \...-. 
                      ,'           | 
                    ,'             | 
                 _,'---    -.      `-._ 
               ,'            `".       ) 
            ,-'                 \     / 
        ,..'-                    `.  (           _ 
       (     ,--._     ,--._      `\  \        ,' |          
      ,-'   _...._\ _,.._   \      `\  )      /   \__ 
     /    ,'      ;'     `.          \/     ,/       ) 
    |    |       |         \          `.   ,'.     ,' 
    | O  :    O  |      O  |           '"--'  \   ( 
    `._  |       \         '                  |    \ 
     . --`.      _`-.___,,'                  ,'     ) 
   .-      `'--.'  `-.__,,-'                ,'   ,-' 
 ,'        '--.-'    .                  _       ( 
(                   _.`--.-        `-..' `.      `. 
 \             _,--'                  )   |   ,.--' 
  '-.......---' 'v'""".      _Y      /    `.  | 
                (_     ;----' |     (      `.,' 
                  \   /_)     '-.   _) 
                   `-'          `..' 


*/
